"""Component based on the RGB metalens from the photonics opt testbed.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
from typing import Any, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as onp
from fmmax import basis, fields, fmm, pml, scattering, sources
from jax import tree_util
from scipy import ndimage  # type: ignore[import-untyped]
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base


DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0

EFIELD = "efield"
HFIELD = "hfield"
FIELD_COORDINATES = "field_coordinates"


@dataclasses.dataclass
class MetalensSpec:
    """Defines the physical specification of the a metalens.

    Attributes:
        permittivity_ambient: Permittivity of the ambient material.
        permittivity_metalens: Permittivity of the material comprising the metalens.
        permittivity_substrate: Permittivity of the substrate.
        thickness_ambient: The thickness of the ambient layer, above the focus offset.
        thickness_metalens: The thickness of the metalens.
        thickness_substrate: The thickness of the substrate.
        focus_offset: The distance between the top of the metalens and the focus.
        width_lens: The width of the lens.
        width_pml: Width of the perfectly matched layers at the borders of the
            simulation unit cell.
        pml_lens_offset: The distance between the PML and the lens.
        pml_source_offset: The distance between the PML and the source.
        source_smoothing_fwhm: The full-width of the smoothing filter for the source
            profile.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    permittivity_ambient: complex
    permittivity_metalens: complex
    permittivity_substrate: complex
    thickness_ambient: float
    thickness_lens: float
    thickness_substrate: float

    focus_offset: float

    width_lens: float
    width_pml: float

    pml_lens_offset: float
    pml_source_offset: float
    source_smoothing_fwhm: float

    grid_spacing: float

    @property
    def width(self) -> float:
        """Return the total width of the unit cell."""
        return self.width_lens + 2 * self.width_pml + 2 * self.pml_lens_offset

    @property
    def lens_offset(self) -> float:
        """Return the offset between the edge of the unit cell and the lens."""
        return self.width_pml + self.pml_lens_offset

    @property
    def source_offset(self) -> float:
        """Return the offset between the edge of the unit cell and the source."""
        return self.width_pml + self.pml_source_offset

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        with jax.ensure_compile_time_eval():
            return (int(jnp.ceil(self.width / self.grid_spacing)), 1)


@dataclasses.dataclass
class MetalensSimParams:
    """Parameters that configure the simulation of a metalens.

    Attributes:
        wavelength: The wavelength of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        num_layers: The number of layers used to discretize the metalens in the
            z-direction.
    """

    wavelength: jnp.ndarray
    formulation: fmm.Formulation
    approximate_num_terms: int
    num_layers: int


@dataclasses.dataclass
class MetalensResponse:
    """Contains the response of the metalens.

    Attributes:
        wavelength: The wavelengths of the metalens simulation.
        enhancement_ex: The enhancement of intensity at the focal point for an
            excitation with x-polarized electric field.
        enhancement_ey: The enhancement of intensity at the focal point for an
            excitation with y-polarized electric field.
    """

    wavelength: jnp.ndarray
    enhancement_ex: jnp.ndarray
    enhancement_ey: jnp.ndarray


json_utils.register_custom_type(MetalensResponse)

tree_util.register_pytree_node(
    MetalensResponse,
    lambda r: (
        (
            r.wavelength,
            r.enhancement_ex,
            r.enhancement_ey,
        ),
        None,
    ),
    lambda _, children: MetalensResponse(*children),
)

# -----------------------------------------------------------------------------
# Define the metalens component.
# -----------------------------------------------------------------------------


@dataclasses.dataclass
class MetalensComponent(base.Component):
    """A metalens component whose only optimizable parameter is density."""

    def __init__(
        self,
        spec: MetalensSpec,
        sim_params: MetalensSimParams,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the metalens components.

        Args:
            spec: Defines the physical specification of the metalens.
            sim_params: Defines simulation parameters for the metalens.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.seed_density = seed_density(spec=self.spec, **seed_density_kwargs)
        self.density_initializer = density_initializer

        # Manually create the expansion.
        nmax = sim_params.approximate_num_terms // 2
        ix = onp.zeros((2 * nmax + 1,), dtype=int)
        ix[1::2] = -onp.arange(1, nmax + 1, dtype=int)
        ix[2::2] = onp.arange(1, nmax + 1, dtype=int)
        assert tuple(ix[:5].tolist()) == (0, -1, 1, -2, 2)
        self.expansion = basis.Expansion(
            basis_coefficients=onp.stack([ix, onp.zeros_like(ix)], axis=-1)
        )

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the metalens component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[basis.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[MetalensResponse, base.AuxDict]:
        """Computes the response of the metalens.

        The response consists of the of the enhancement of fields at the focal point
        for incident plane wave having x- and y-polarized electric fields.

        Args:
            params: The parameters defining the metalens, matching those returned
                by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, the `aux` will include an x-z cross section
                of the electric and magnetic fields.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        return simulate_metalens(
            density=params,
            spec=self.spec,
            wavelength=jnp.asarray(wavelength),
            expansion=expansion,
            num_layers=self.sim_params.num_layers,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )


def seed_density(spec: MetalensSpec, **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a metalens component.

    Args:
        spec: The specification of the metalens component.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a metalens component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2

    density_shape = (
        spec.grid_shape[0],
        int(jnp.round(spec.thickness_lens / spec.grid_spacing)),
    )
    fixed_solid = onp.zeros(density_shape, dtype=bool)
    fixed_void = onp.zeros(density_shape, dtype=bool)
    fixed_void[:, 0] = True  # Top, adjacent to ambient.
    delta = int(jnp.round(spec.lens_offset / spec.grid_spacing))
    fixed_void[:delta, :-1] = True
    fixed_void[-delta:, :-1] = True
    fixed_solid[:, -1] = True  # Bottom, adjacent to substrate
    return types.Density2DArray(
        array=jnp.full(density_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        fixed_solid=fixed_solid,
        fixed_void=fixed_void,
        periodic=(False, False),
        **kwargs,
    )


# -----------------------------------------------------------------------------
# Simulation method for the metalens.
# -----------------------------------------------------------------------------


def simulate_metalens(
    density: types.Density2DArray,
    spec: MetalensSpec,
    wavelength: jnp.ndarray,
    expansion: basis.Expansion,
    num_layers: int,
    formulation: fmm.Formulation,
    compute_fields: bool,
) -> Tuple[MetalensResponse, base.AuxDict]:
    """Simulates the metalens component.

    Args:
        density: Defines the pattern of the metalens.
        spec: Defines the physical specifcation of the metalens.
        wavelength: The wavelength of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        num_layers: The number of layers used to discretize the metalens in the
            z-direction.
        formulation: Defines the FMM formulation to be used.
        compute_fields: If `True`, returns electric and magnetic fields in the
            `aux` dictionary.

    Returns:
        The `ExtractorResponse` and `aux` dictionary.
    """
    density_array = utils.transforms.rescaled_density_array(
        density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    assert spec.grid_shape == (density_array.shape[0], 1)
    dim = spec.grid_shape[0]

    # We accommodate density for metalenses having thickness different than that
    # specified in the `spec`. When the thickness is greater or lesser, the focus
    # offset (normally measured from the top surface of the lens) as adjusted to
    # compensate.
    #
    # Also, trim the top and bottom layer from the density, which should be entirely
    # void (ambient material) and entirely solid (substrate material), respectively.
    # Add the single-pixel thickness to the focus offset and substrate so the total
    # stack height remains unchanged. This allows the discretization of the metalens
    # in the z-direction to be a bit more accurate.
    density_array = density_array[:, 1:-1]
    thickness_lens = density_array.shape[-1] * spec.grid_spacing
    focus_offset = spec.focus_offset + spec.grid_spacing
    thickness_substrate = spec.thickness_substrate + spec.grid_spacing

    layer_thicknesses = (
        [jnp.asarray(focus_offset + spec.thickness_ambient)]
        + [jnp.asarray(thickness_lens / num_layers)] * num_layers
        + [jnp.asarray(thickness_substrate)]
    )

    # Split the metalens density into `num_layers` slices.
    density_array = utils.transforms.resample(
        density_array, shape=(density_array.shape[0], num_layers)
    )
    metalens_densities = jnp.split(density_array, list(range(1, num_layers)), axis=-1)
    assert len(metalens_densities) == num_layers
    assert all([d.shape == spec.grid_shape for d in metalens_densities])
    metalens_permittivities = [
        utils.transforms.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_metalens),
            permittivity_void=jnp.asarray(spec.permittivity_ambient),
            density=d,
        )
        for d in metalens_densities
    ]

    in_plane_wavevector = jnp.zeros((2,))
    primitive_lattice_vectors = basis.LatticeVectors(
        u=spec.width * basis.X,
        v=basis.Y,
    )

    def eigensolve_fn(permittivity: jnp.ndarray) -> fmm.LayerSolveResult:
        # Permittivities and permeabilities are returned in the order needed
        # for the anisotropic eigensolve below.
        permittivities_pml, permeabilities_pml = pml.apply_uniaxial_pml(
            permittivity=permittivity,
            pml_params=pml.PMLParams(
                num_x=int(spec.width_pml / spec.grid_spacing),
                num_y=0,
            ),
        )
        return fmm.eigensolve_general_anisotropic_media(
            wavelength,
            in_plane_wavevector,
            primitive_lattice_vectors,
            *permittivities_pml,
            *permeabilities_pml,
            expansion=expansion,
            formulation=formulation,
            vector_field_source=jnp.mean(jnp.asarray(permittivities_pml), axis=0),
        )

    with jax.ensure_compile_time_eval():
        solve_result_ambient = eigensolve_fn(
            permittivity=jnp.full(spec.grid_shape, spec.permittivity_ambient)
        )
        solve_result_substrate = eigensolve_fn(
            permittivity=jnp.full(spec.grid_shape, spec.permittivity_substrate)
        )

    if compute_fields:
        # If the field calculation is desired, compute the interior scattering
        # matrices. For each layer in the stack, the interior scattering matrices
        # consist of a pair of matrices, one for the substack below the layer, and
        # one for the substack above the layer.
        solve_results_metalens = [
            eigensolve_fn(permittivity=p) for p in metalens_permittivities
        ]
        layer_solve_results = (
            [solve_result_ambient] + solve_results_metalens + [solve_result_substrate]
        )
        s_matrices_interior = scattering.stack_s_matrices_interior(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )
        s_matrix = s_matrices_interior[-1][0]
    else:
        solve_results_metalens_batch = eigensolve_fn(
            permittivity=jnp.asarray(metalens_permittivities)[:, jnp.newaxis, :, :]
        )
        # Merge with the ambient and substrate solve results to get the solve results
        # for the full stack, needed by `stack_s_matrix_scan`.
        stack_layer_solve_results = tree_util.tree_map(
            lambda a, b, c: jnp.concatenate(
                [
                    a[jnp.newaxis, ...],
                    jnp.broadcast_to(b, (num_layers,) + b.shape[1:]),
                    c[jnp.newaxis, ...],
                ]
            ),
            solve_result_ambient,
            solve_results_metalens_batch,
            solve_result_substrate,
        )
        s_matrix = scattering.stack_s_matrix_scan(
            layer_solve_results=stack_layer_solve_results,
            layer_thicknesses=jnp.asarray(layer_thicknesses),
        )

    # Compute the source, consisting of a smoothed step function.
    # TODO: a jax implementation of the smoothing operation.
    with jax.ensure_compile_time_eval():
        x = onp.arange(dim) / dim * spec.width
        profile = (x > spec.source_offset) & (x < spec.width - spec.source_offset)
        profile = profile.astype(float)
        sigma = spec.source_smoothing_fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))
        sigma_pixels = sigma / spec.grid_spacing
        profile = ndimage.gaussian_filter1d(profile, sigma=sigma_pixels)
        profile = profile[:, onp.newaxis]
        norm = 1 / jnp.sqrt(spec.permittivity_substrate)
        incident_ex = jnp.stack([-profile, jnp.zeros_like(profile)], axis=-1) * norm
        incident_ey = jnp.stack([jnp.zeros_like(profile), profile], axis=-1) * norm
        incident_hx = jnp.stack([jnp.zeros_like(profile), profile], axis=-1)
        incident_hy = jnp.stack([profile, jnp.zeros_like(profile)], axis=-1)

        _, bwd_amplitude_substrate_end = sources.amplitudes_for_fields(
            ex=incident_ex,
            ey=incident_ey,
            hx=incident_hx,
            hy=incident_hy,
            layer_solve_result=solve_result_substrate,
            brillouin_grid_axes=None,
        )

    def compute_intensity_at_focus(s_matrix):
        # Compute the field at the focal point.
        bwd_amplitude_ambient_end = s_matrix.s22 @ bwd_amplitude_substrate_end
        bwd_amplitude_ambient_focus = fields.propagate_amplitude(
            bwd_amplitude_ambient_end,
            distance=focus_offset,
            layer_solve_result=solve_result_ambient,
        )
        ef_focus, hf_focus = fields.fields_from_wave_amplitudes(
            forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_focus),
            backward_amplitude=bwd_amplitude_ambient_focus,
            layer_solve_result=solve_result_ambient,
        )
        (ex_focus, ey_focus, ez_focus), _, _ = fields.fields_on_coordinates(
            electric_field=ef_focus,
            magnetic_field=hf_focus,
            layer_solve_result=solve_result_ambient,
            x=jnp.asarray(spec.width / 2),
            y=jnp.asarray(0.0),
        )
        return jnp.abs(ex_focus) ** 2 + jnp.abs(ey_focus) ** 2 + jnp.abs(ez_focus) ** 2

    # Compute the intensity at the target focal point in the absence of a lens.
    # This is needed to compute the intensity enhancement due to the lens.
    with jax.ensure_compile_time_eval():
        s_matrix_no_lens = scattering.stack_s_matrix(
            layer_solve_results=[solve_result_ambient, solve_result_substrate],
            layer_thicknesses=[jnp.zeros(()), jnp.zeros(())],
        )
        intensity_no_lens = compute_intensity_at_focus(s_matrix_no_lens)

    intensity = compute_intensity_at_focus(s_matrix)
    enhancement = intensity / intensity_no_lens
    assert enhancement.shape == solve_result_ambient.batch_shape + (2,)
    enhancement_ex = enhancement[..., 0]
    enhancement_ey = enhancement[..., 1]
    response = MetalensResponse(
        wavelength=solve_result_ambient.wavelength,
        enhancement_ex=enhancement_ex,
        enhancement_ey=enhancement_ey,
    )

    # -------------------------------------------------------------------------
    # Compute fields in an xz cross section.
    # -------------------------------------------------------------------------

    aux = {}
    if compute_fields:
        amplitudes_interior = fields.stack_amplitudes_interior(
            s_matrices_interior=s_matrices_interior,
            forward_amplitude_0_start=jnp.zeros_like(bwd_amplitude_substrate_end),
            backward_amplitude_N_end=bwd_amplitude_substrate_end,
        )
        layer_znum = tuple(
            [int(jnp.round(t / spec.grid_spacing) + 1) for t in layer_thicknesses]
        )
        (ex, ey, ez), (hx, hy, hz), (xf, yf, zf) = fields.stack_fields_3d(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
            layer_znum=layer_znum,
            grid_shape=spec.grid_shape,
            num_unit_cells=(1, 1),
        )
        aux.update(
            {
                EFIELD: (ex, ey, ez),
                HFIELD: (hx, hy, hz),
                FIELD_COORDINATES: (xf, yf, zf),
            }
        )

    return response, aux



"""Defines the metagrating challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Sequence, Tuple

import jax.numpy as jnp
from fmmax import basis, fmm  # type: ignore[import-untyped]
from jax import nn
from totypes import symmetry, types

from invrs_gym.challenges import base
from invrs_gym.challenges.diffract import common
from invrs_gym.utils import initializers

AVERAGE_EFFICIENCY = "average_efficiency"
MIN_EFFICIENCY = "min_efficiency"

POLARIZATION = "TM"

density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class MetagratingChallenge(base.Challenge):
    """Defines the metagrating challenge.

    The objective of the metagrating challenge is to design a density so that incident
    light is efficiently diffracted into the specified transmission order.

    Attributes:
        component: The component to be designed.
        transmission_order: The transmission diffraction order to be maximized.
    """

    component: common.SimpleGratingComponent
    transmission_order: Tuple[int, int]

    def loss(
        self,
        response: common.GratingResponse,
        transmission_target: float = 0.95,
    ) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`.

        Args:
            response: The response obtained from simulating the component.
            transmission_target: Value used in a nonlinearity within the loss function.
                When the transmission value is far from the target, the dependence of
                loss on tranmsission is larger.

        Returns:
            The scalar loss value.
        """
        # Compute efficiency, a per-wavelength scalar.
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
            polarization=POLARIZATION,
        )
        batch_shape = jnp.broadcast_shapes(
            response.wavelength.shape,
            response.polar_angle.shape,
            response.azimuthal_angle.shape,
        )
        assert efficiency.shape == batch_shape
        window_size = 1 - transmission_target
        scaled_error = (transmission_target - efficiency) / window_size
        return jnp.mean(nn.softplus(scaled_error) ** 2)

    def eval_metric(self, response: common.GratingResponse) -> jnp.ndarray:
        """Compute eval metric from the component `response`.

        The eval metric is the minimum efficiency of diffraction into the target order
        across all wavelengths.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
            polarization=POLARIZATION,
        )
        return jnp.amin(efficiency)

    def metrics(
        self,
        response: common.GratingResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        metrics = super().metrics(response, params, aux)
        efficiency = _value_for_order(
            response.transmission_efficiency,
            expansion=response.expansion,
            order=self.transmission_order,
            polarization=POLARIZATION,
        )
        metrics.update(
            {
                AVERAGE_EFFICIENCY: jnp.mean(efficiency),
                MIN_EFFICIENCY: jnp.amin(efficiency),
            }
        )
        return metrics


def _value_for_order(
    array: jnp.ndarray,
    expansion: basis.Expansion,
    order: Tuple[int, int],
    polarization: str,
) -> jnp.ndarray:
    """Extracts the value from `array` for the specified Fourier order."""
    assert polarization in ("TE", "TM")
    polarization_idx = 0 if polarization == "TE" else 1
    order_idx = common.index_for_order(order, expansion)
    return array[..., order_idx, polarization_idx]


# -----------------------------------------------------------------------------
# Metagrating with 1.371 x 0.525 um design region.
# -----------------------------------------------------------------------------


METAGRATING_SPEC = common.GratingSpec(
    permittivity_ambient=(1.0 + 0.0j) ** 2,
    permittivity_cap=(1.0 + 0.0j) ** 2,
    permittivity_grating=(3.45 + 0.00001j) ** 2,
    permittivity_encapsulation=(1.0 + 0.00001j) ** 2,
    permittivity_spacer=(1.45 + 0.0j) ** 2,
    permittivity_substrate=(1.45 + 0.0j) ** 2,
    thickness_cap=0.0,
    thickness_grating=0.325,
    thickness_spacer=0.0,
    period_x=float(1.050 / jnp.sin(jnp.deg2rad(50.0))),
    period_y=0.525,
    grid_spacing=0.0117,  # Yields a grid shape of `(118, 45)`.
)

METAGRATING_SIM_PARAMS = common.GratingSimParams(
    wavelength=1.050,
    polar_angle=0.0,
    azimuthal_angle=0.0,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    approximate_num_terms=300,
    truncation=basis.Truncation.CIRCULAR,
)

SYMMETRIES = (symmetry.REFLECTION_E_W,)

# Objective is to diffract light into the +1 transmitted order.
TRANSMISSION_ORDER = (1, 0)


def metagrating(
    minimum_width: int = 5,
    minimum_spacing: int = 5,
    density_initializer: base.DensityInitializer = density_initializer,
    transmission_order: Tuple[int, int] = TRANSMISSION_ORDER,
    spec: common.GratingSpec = METAGRATING_SPEC,
    sim_params: common.GratingSimParams = METAGRATING_SIM_PARAMS,
    symmetries: Sequence[str] = SYMMETRIES,
) -> MetagratingChallenge:
    """Metagrating challenge with 1.371 x 0.525 um design region.

    The metagrating challenge is based on the metagrating example in "Validation and
    characterization of algorithms for photonics inverse design" by Chen et al.
    (in preparation), in which designs with feature size from 35 to 66 nm are shown.

    It involves maximizing diffraction of light transmitted from a silicon oxide
    substrate into the ambient using a patterned silicon metastructure. The excitation
    is TM-polarized plane wave with 1.05 micron wavelength.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            default value of 5 corresponds to a physical size of approximately 60 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.
        transmission_order: The diffraction order to be maximized.
        spec: Defines the physical specification of the metagrating.
        sim_params: Defines the simulation settings of the metagrating.
        symmetries: Defines the symmetries of the metagrating.

    Returns:
        The `MetagratingChallenge`.
    """
    return MetagratingChallenge(
        component=common.SimpleGratingComponent(
            spec=spec,
            sim_params=sim_params,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
        transmission_order=transmission_order,
    )



"""Defines initializers for densities and other objects.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Any

import jax
import jax.numpy as jnp
from totypes import types


def identity_initializer(key: jax.Array, seed_obj: Any) -> Any:
    """Returns `seed_obj`."""
    del key
    return seed_obj


def noisy_density_initializer(
    key: jax.Array,
    seed_density: types.Density2DArray,
    relative_mean: jnp.ndarray | float,
    relative_noise_amplitude: jnp.ndarray | float,
    resize_method: jax.image.ResizeMethod = jax.image.ResizeMethod.CUBIC,
) -> types.Density2DArray:
    """Return a density with specified mean and added random noise.

    Only metadata from the seed density (e.g. feature sizes) are retained; the array
    for the output density is randomly generated. It has the specified mean, and added
    uniform random noise with the specified amplitude and a length scale determined
    by the minimum width and spacing of the `seed_density`.

    Args:
        key: Key used in the generation of random noise.
        seed_density: The density used to provide metadata.
        relative_mean: The relative mean value of the output density. For a value of
            `0.5`, the mean is between the density upper and lower bounds.
        relative_noise_amplitude: The relative amplitude of noise added to the mean.
        resize_method: The method used to resize a low-resolution array to the final
            array, ensuring the length scale of added noise.

    Returns:
        The `Density2DArray` with added noise.
    """
    mean = (
        seed_density.lower_bound
        + (seed_density.upper_bound - seed_density.lower_bound) * relative_mean
    )
    array = jnp.full(seed_density.shape, mean)

    length_scale = (seed_density.minimum_spacing + seed_density.minimum_width) / 2
    low_res_shape = seed_density.shape[:-2] + tuple(
        int(jnp.ceil(s / length_scale)) for s in seed_density.shape[-2:]
    )
    low_res_noise = jax.random.uniform(key, low_res_shape) - 0.5
    noise = jax.image.resize(low_res_noise, seed_density.shape, method=resize_method)
    noise *= relative_noise_amplitude * (
        seed_density.upper_bound - seed_density.lower_bound
    )

    array += noise
    array = jnp.clip(array, seed_density.lower_bound, seed_density.upper_bound)

    density = substitute_array(array, seed_density)
    return types.symmetrize_density(density)


# -----------------------------------------------------------------------------
# Functions used in density initializers.
# -----------------------------------------------------------------------------


def apply_fixed_pixels(density: types.Density2DArray) -> types.Density2DArray:
    """Return a density with fixed void and solid pixels set to respective bounds."""
    arr = density.array
    if density.fixed_solid is not None:
        arr = jnp.where(density.fixed_solid, density.upper_bound, arr)
    if density.fixed_void is not None:
        arr = jnp.where(density.fixed_void, density.lower_bound, arr)
    return substitute_array(jnp.asarray(arr), density)


def substitute_array(
    arr: jnp.ndarray, density: types.Density2DArray
) -> types.Density2DArray:
    """Return a new density identical to `density` but with `arr` as the `array`."""
    assert arr.shape == density.shape
    return jax.tree_util.tree_unflatten(jax.tree_util.tree_structure(density), [arr])



"""Defines various transformation functions.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Tuple

import jax
from jax import numpy as jnp
from totypes import types


def rescaled_density_array(
    density: types.Density2DArray,
    lower_bound: float,
    upper_bound: float,
) -> jnp.ndarray:
    """Return a density array for specified lower and upper bounds."""
    array = density.array - density.lower_bound
    array /= density.upper_bound - density.lower_bound
    array *= upper_bound - lower_bound
    return jnp.asarray(array + lower_bound)


def interpolate_permittivity(
    permittivity_solid: jnp.ndarray,
    permittivity_void: jnp.ndarray,
    density: jnp.ndarray,
) -> jnp.ndarray:
    """Interpolates the permittivity with a scheme that avoids zero crossings.

    The interpolation uses the scheme introduced in [2019 Christiansen], which avoids
    zero crossings that can occur with metals or lossy materials having a negative
    real component of the permittivity. https://doi.org/10.1016/j.cma.2018.08.034

    Args:
        permittivity_solid: The permittivity of solid regions.
        permittivity_void: The permittivity of void regions.
        density: The density, specifying which locations are solid and which are void.

    Returns:
        The interpolated permittivity.
    """
    n_solid = jnp.real(jnp.sqrt(permittivity_solid))
    k_solid = jnp.imag(jnp.sqrt(permittivity_solid))
    n_void = jnp.real(jnp.sqrt(permittivity_void))
    k_void = jnp.imag(jnp.sqrt(permittivity_void))
    n = density * n_solid + (1 - density) * n_void
    k = density * k_solid + (1 - density) * k_void
    return (n + 1j * k) ** 2


def resample(
    x: jnp.ndarray,
    shape: Tuple[int, ...],
    method: jax.image.ResizeMethod = jax.image.ResizeMethod.LINEAR,
) -> jnp.ndarray:
    """Resamples `x` to have the specified `shape`.

    The algorithm first upsamples `x` so that the pixels in the output image are
    comprised of an integer number of pixels in the upsampled `x`, and then
    performs box downsampling.

    Args:
        x: The array to be resampled.
        shape: The shape of the output array.
        method: The method used to resize `x` prior to box downsampling.

    Returns:
        The resampled array.
    """
    if x.ndim != len(shape):
        raise ValueError(
            f"`shape` must have length matching number of dimensions in `x`, "
            f"but got {shape} when `x` had shape {x.shape}."
        )

    with jax.ensure_compile_time_eval():
        factor = [int(jnp.ceil(dx / d)) for dx, d in zip(x.shape, shape)]
        upsampled_shape = tuple([d * f for d, f in zip(shape, factor)])

    x_upsampled = jax.image.resize(
        image=x,
        shape=upsampled_shape,
        method=method,
    )

    return box_downsample(x_upsampled, shape)


def box_downsample(x: jnp.ndarray, shape: Tuple[int, ...]) -> jnp.ndarray:
    """Downsamples `x` to a coarser resolution array using box downsampling.

    Box downsampling forms nonoverlapping windows and simply averages the
    pixels within each window. For example, downsampling `(0, 1, 2, 3, 4, 5)`
    with a factor of `2` yields `(0.5, 2.5, 4.5)`.

    Args:
        x: The array to be downsampled.
        shape: The shape of the output array; each axis dimension must evenly
            divide the corresponding axis dimension in `x`.

    Returns:
        The output array with shape `shape`.
    """
    if x.ndim != len(shape) or any([(d % s) != 0 for d, s in zip(x.shape, shape)]):
        raise ValueError(
            f"Each axis of `shape` must evenly divide the corresponding axis "
            f"dimension in `x`, but got shape {shape} when `x` has shape "
            f"{x.shape}."
        )
    shape = sum([(s, d // s) for d, s in zip(x.shape, shape)], ())
    axes = list(range(1, 2 * x.ndim, 2))
    x = x.reshape(shape)
    return jnp.mean(x, axis=axes)



"""Defines functions common across diffract challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
from typing import Any, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as onp
from fmmax import basis, fields, fmm, scattering  # type: ignore[import-untyped]
from jax import tree_util
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base

Params = Any

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0

THICKNESS_CAP = "thickness_cap"
THICKNESS_GRATING = "thickness_grating"
THICKNESS_SPACER = "thickness_spacer"
DENSITY = "density"


@dataclasses.dataclass
class GratingSpec:
    """Defines the physical specifcation of a grating.

    Attributes:
        permittivity_ambient: Permittivity of the ambient material.
        permittivity_cap: Permittivity of the material between grating and ambient.
        permittivity_grating: Permittivity of the grating teeth.
        permittivity_encapsulation: Permittivity of the material in gaps between
            grating teeth.
        permittivity_spacer: Permittivity of the spacer layer between grating and
            substrate.
        permittivity_substrate: Permittivity of the substrate.
        thickness_cap: Thickness of the cap layer between grating and ambient.
        thickness_grating: Thickness of the grating layer.
        thickness_spacer: Thickness of the spacer layer between grating and substrate.
        period_x: The size of the unit cell along the x direction.
        period_y: The size of the unit cell along the y direction.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    permittivity_ambient: complex
    permittivity_cap: complex
    permittivity_grating: complex
    permittivity_encapsulation: complex
    permittivity_spacer: complex
    permittivity_substrate: complex

    thickness_cap: float | jnp.ndarray | types.BoundedArray
    thickness_grating: float | jnp.ndarray | types.BoundedArray
    thickness_spacer: float | jnp.ndarray | types.BoundedArray

    period_x: float
    period_y: float

    grid_spacing: float

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        return (
            int(jnp.ceil(self.period_x / self.grid_spacing)),
            int(jnp.ceil(self.period_y / self.grid_spacing)),
        )


@dataclasses.dataclass
class GratingSimParams:
    """Parameters that configure the simulation of a grating.

    Attributes:
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    wavelength: float | jnp.ndarray
    polar_angle: float | jnp.ndarray
    azimuthal_angle: float | jnp.ndarray
    formulation: fmm.Formulation
    approximate_num_terms: int
    truncation: basis.Truncation


@dataclasses.dataclass
class GratingResponse:
    """Contains the response of the grating.

    Attributes:
        wavelength: The wavelength for the efficiency calculation.
        polar_angle: The polar angle for the efficiency calculation.
        azimuthal_angle: The azimuthal angle for the efficiency calculation.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is transmitted.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is reflected.
        expansion: Defines the Fourier expansion for the calculation.
    """

    wavelength: jnp.ndarray
    polar_angle: jnp.ndarray
    azimuthal_angle: jnp.ndarray
    transmission_efficiency: jnp.ndarray
    reflection_efficiency: jnp.ndarray
    expansion: basis.Expansion


json_utils.register_custom_type(GratingResponse)

tree_util.register_pytree_node(
    GratingResponse,
    lambda r: (
        (
            r.wavelength,
            r.polar_angle,
            r.azimuthal_angle,
            r.transmission_efficiency,
            r.reflection_efficiency,
            r.expansion,
        ),
        None,
    ),
    lambda _, children: GratingResponse(*children),
)


# -----------------------------------------------------------------------------
# Define components used in diffract optimization challenges.
# -----------------------------------------------------------------------------


class SimpleGratingComponent(base.Component):
    """A simple grating component whose only optimizable parameter is density."""

    def __init__(
        self,
        spec: GratingSpec,
        sim_params: GratingSimParams,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component.

        Args:
            spec: Defines the physical specification of the grating.
            sim_params: Defines simulation parameters for the grating.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.seed_density = seed_density(
            grid_shape=self.spec.grid_shape,
            **seed_density_kwargs,
        )
        self.density_initializer = density_initializer

        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.period_x * basis.X,
                v=self.spec.period_y * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the grating component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[basis.Expansion] = None,
    ) -> Tuple[GratingResponse, base.AuxDict]:
        """Computes the response of the grating.

        The response consists of the transmitted and reflected power into each order
        for both TE- and TM-polarized plane wave illumination.

        Args:
            params: The parameters defining the metagrating, matching those returned
                by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        transmission_efficiency, reflection_efficiency = grating_efficiency(
            density=params,
            spec=self.spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
        )
        response = GratingResponse(
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, {}


class GratingWithOptimizableThicknessComponent(base.Component):
    """A grating whose optimizable parameters are density and layer thickness."""

    def __init__(
        self,
        spec: GratingSpec,
        sim_params: GratingSimParams,
        thickness_initializer: base.ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component splitter component.

        Args:
            spec: Defines the physical specification of the grating.
            sim_params: Defines simulation parameters for the grating.
            thickness_initializer: Callable which returns the initial thickness for
                the grating layer from a random key and a bounded array with value
                equal the thickness from `spec`.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_density = seed_density(
            grid_shape=self.spec.grid_shape,
            **seed_density_kwargs,
        )

        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.period_x * basis.X,
                v=self.spec.period_y * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for the grating component."""
        keys = iter(jax.random.split(key, num=4))
        params = {
            THICKNESS_CAP: self.thickness_initializer(
                next(keys), self.spec.thickness_cap  # type: ignore[arg-type]
            ),
            THICKNESS_GRATING: self.thickness_initializer(
                next(keys), self.spec.thickness_grating  # type: ignore[arg-type]
            ),
            THICKNESS_SPACER: self.thickness_initializer(
                next(keys), self.spec.thickness_spacer  # type: ignore[arg-type]
            ),
            DENSITY: self.density_initializer(next(keys), self.seed_density),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[basis.Expansion] = None,
    ) -> Tuple[GratingResponse, base.AuxDict]:
        """Computes the response of the grating component.

        The response consists of the transmitted and reflected power into each order
        for both TE- and TM-polarized plane wave illumination.

        Args:
            params: The parameters defining the grating, matching those returned by
                the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        spec = dataclasses.replace(
            self.spec,
            thickness_cap=jnp.asarray(params[THICKNESS_CAP].array),
            thickness_grating=jnp.asarray(params[THICKNESS_GRATING].array),
            thickness_spacer=jnp.asarray(params[THICKNESS_SPACER].array),
        )
        transmission_efficiency, reflection_efficiency = grating_efficiency(
            density=params[DENSITY],  # type: ignore[arg-type]
            spec=spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
        )
        response = GratingResponse(
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, {}


def seed_density(grid_shape: Tuple[int, int], **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a grating component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a grating component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(True, True),
        **kwargs,
    )


def index_for_order(
    order: Tuple[int, int],
    expansion: basis.Expansion,
) -> int:
    """Returns the index for the specified Fourier order and expansion."""
    ((order_idx,),) = onp.where(onp.all(expansion.basis_coefficients == order, axis=1))
    assert tuple(expansion.basis_coefficients[order_idx, :]) == order
    return int(order_idx)


# -----------------------------------------------------------------------------
# Simulation method used by all gratings.
# -----------------------------------------------------------------------------


def grating_efficiency(
    density: types.Density2DArray,
    spec: GratingSpec,
    wavelength: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Compute the per-order transmission and reflection efficiency for a grating.

    The excitation for the calculation are separate a TE- and TM-polarized plane waves
    at the specified wavelength(s), incident from the substrate with the specified
    polar and azimuthal angles.

    Args:
        density: Defines the pattern of the grating layer.
        spec: Defines the physical specifcation of the grating.
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.

    Returns:
        The per-order transmission and reflection efficiency, having shape
        `(num_wavelengths, num_fourier_terms, 2)`.
    """
    density_array = utils.transforms.rescaled_density_array(
        density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    permittivities = (
        jnp.full((1, 1), spec.permittivity_ambient),
        jnp.full((1, 1), spec.permittivity_cap),
        utils.transforms.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_grating),
            permittivity_void=jnp.asarray(spec.permittivity_encapsulation),
            density=density_array,
        ),
        jnp.full((1, 1), spec.permittivity_spacer),
        jnp.full((1, 1), spec.permittivity_substrate),
    )

    in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        # Polar angle is defined in substrate, since light is incident from substrate.
        permittivity=jnp.asarray(spec.permittivity_substrate),
    )
    layer_solve_results = [
        fmm.eigensolve_isotropic_media(
            wavelength=jnp.asarray(wavelength),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=basis.LatticeVectors(
                u=spec.period_x * basis.X,
                v=spec.period_y * basis.Y,
            ),
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]

    # Layer thicknesses for the ambient and substrate are set to zero; these do not
    # affect the result of the calculation.
    layer_thicknesses = (
        jnp.zeros(()),
        jnp.asarray(spec.thickness_cap),
        jnp.asarray(spec.thickness_grating),
        jnp.asarray(spec.thickness_spacer),
        jnp.zeros(()),
    )

    s_matrix = scattering.stack_s_matrix(layer_solve_results, layer_thicknesses)

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate the wave amplitudes for backward-going TE or TM-polarized plane waves
    # at the end of substrate layer.
    bwd_amplitude_substrate_end = jnp.zeros((2 * n, 2), dtype=complex)
    bwd_amplitude_substrate_end = bwd_amplitude_substrate_end.at[0, 0].set(1.0)  # TE
    bwd_amplitude_substrate_end = bwd_amplitude_substrate_end.at[n, 1].set(1.0)  # TM

    # Calculate the incident power from the substrate. Since the substrate thickness
    # has been set to zero, the forward and backward amplitudes are already colocated.
    fwd_amplitude_substrate_start = s_matrix.s12 @ bwd_amplitude_substrate_end
    fwd_flux_substrate, bwd_flux_substrate = fields.amplitude_poynting_flux(
        forward_amplitude=fwd_amplitude_substrate_start,
        backward_amplitude=bwd_amplitude_substrate_end,
        layer_solve_result=layer_solve_results[-1],
    )

    # Sum over orders and polarizations to get the total incident flux.
    total_incident_flux = jnp.sum(bwd_flux_substrate, axis=-2, keepdims=True)

    # Calculate the transmitted power in the ambient.
    bwd_amplitude_ambient_end = s_matrix.s22 @ bwd_amplitude_substrate_end
    _, bwd_flux_ambient = fields.amplitude_poynting_flux(
        forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
        backward_amplitude=bwd_amplitude_ambient_end,
        layer_solve_result=layer_solve_results[0],
    )

    # Sum the fluxes over the two polarizations for each order. Note that the sum is
    # done for each excitation condition (i.e. TE and TM), so as to capture the
    # effect of polarization conversion.
    bwd_flux_ambient = bwd_flux_ambient[..., :n, :] + bwd_flux_ambient[..., n:, :]
    fwd_flux_substrate = fwd_flux_substrate[..., :n, :] + fwd_flux_substrate[..., n:, :]

    transmission_efficiency = bwd_flux_ambient / total_incident_flux
    reflection_efficiency = fwd_flux_substrate / total_incident_flux

    return transmission_efficiency, reflection_efficiency



"""Defines the ceviche challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Any, Optional, Sequence, Tuple

import ceviche_challenges as cc  # type: ignore[import-untyped]
import ceviche_challenges.wdm.model as wdm_model  # type: ignore[import-untyped]
import jax
import jax.numpy as jnp
import numpy as onp
from agjax.experimental import wrapper
from ceviche_challenges import units as u  # type: ignore[import-untyped]
from jax import tree_util
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base
from invrs_gym.challenges.ceviche import defaults
from invrs_gym.loss import transmission_loss

Params = Any

FIELDS = "fields"
SPARAMS = "sparams"

TRANSMISSION_EXPONENT = 1.0
SCALAR_EXPONENT = 2.0

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0


density_initializer = functools.partial(
    utils.initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


class CevicheComponent(base.Component):
    """Defines a general ceviche component."""

    def __init__(
        self,
        ceviche_model: defaults.Model,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initialize a `CevicheComponent`.

        Args:
            ceviche_model: The model for the component.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """
        self.ceviche_model = ceviche_model
        self.seed_density = _seed_density(ceviche_model, **seed_density_kwargs)
        self.density_initializer = density_initializer

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        excite_port_idxs: Sequence[int] = (0,),
        wavelengths_nm: Optional[jnp.ndarray] = None,
        max_parallelizm: Optional[int] = 1,
    ) -> Tuple["CevicheResponse", base.AuxDict]:
        """Compute the response of the component and auxilliary quantities."""

        with jax.ensure_compile_time_eval():
            if wavelengths_nm is None:
                wavelengths_nm = jnp.asarray(self.ceviche_model.output_wavelengths)

        # The ceviche simulation function is autograd-differentiable. Wrap it so that
        # it can be differentiated using jax.
        autograd_sim_fn = functools.partial(
            self.ceviche_model.simulate,
            excite_port_idxs=excite_port_idxs,
            wavelengths_nm=wavelengths_nm,
            max_parallelizm=max_parallelizm,
        )

        s_params_shape = (
            len(wavelengths_nm),
            len(excite_port_idxs),
            len(self.ceviche_model.ports),
        )
        fields_shape = s_params_shape[:2] + self.ceviche_model.shape
        sim_fn = wrapper.wrap_for_jax(
            autograd_sim_fn,
            result_shape_dtypes=(
                jnp.ones(s_params_shape, dtype=complex),
                jnp.ones(fields_shape, dtype=complex),
            ),
            nondiff_outputnums=(1,),
        )

        density_array = utils.transforms.rescaled_density_array(
            params,
            lower_bound=DENSITY_LOWER_BOUND,
            upper_bound=DENSITY_UPPER_BOUND,
        )
        sparams, fields = sim_fn(density_array)
        response = CevicheResponse(
            s_parameters=sparams,
            wavelengths_nm=wavelengths_nm,
            excite_port_idxs=jnp.asarray(excite_port_idxs),
        )
        aux = {SPARAMS: sparams, FIELDS: fields}
        return response, aux


@dataclasses.dataclass
class CevicheResponse:
    """Stores the response of a Ceviche component."""

    s_parameters: jnp.ndarray
    wavelengths_nm: jnp.ndarray
    excite_port_idxs: jnp.ndarray


tree_util.register_pytree_node(
    CevicheResponse,
    flatten_func=lambda r: ((r.s_parameters, r.wavelengths_nm, r.excite_port_idxs), ()),
    unflatten_func=lambda aux, children: CevicheResponse(*children, *aux),
)


json_utils.register_custom_type(CevicheResponse)


def _seed_density(ceviche_model: defaults.Model, **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for the `ceviche_model`.

    The seed density has shape and fixed pixels as required by the `ceviche_model`,
    and with other properties determined by keyword arguments.

    Args:
        ceviche_model: The model from which the seed density determined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are computed from `ceviche_model`, which must not be specified.
    invalid_kwargs = (
        "array",
        "fixed_solid",
        "fixed_void",
        "lower_bound",
        "upper_bound",
    )
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    shape = ceviche_model.design_variable_shape
    fixed_solid, fixed_void = _fixed_pixels(ceviche_model)
    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    seed_density = types.Density2DArray(
        array=jnp.full(shape, mid_density_value),
        fixed_solid=fixed_solid,
        fixed_void=fixed_void,
        # For the ceviche challenges, density must lie between 0 and 1.
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        **kwargs,
    )
    return seed_density


def _fixed_pixels(ceviche_model: defaults.Model) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Determine the fixed pixels for the given ceviche model.

    The fixed pixels exist around the border of the design, and are solid where the
    adjacent pixels outside the design are solid. They are void where the adjacent
    pixels outside the design are void.

    Args:
        ceviche_model: The model from which the fixed pixels are determined.

    Returns:
        The fixed solid and void pixels.
    """
    design_shape = ceviche_model.design_variable_shape
    density_fn = ceviche_model.density

    assert not onp.any(density_fn(jnp.ones(design_shape)) == 0.5)
    density = density_fn(jnp.full(design_shape, 0.5))

    i, j = onp.where(density == 0.5)
    i_lo = i[0]
    i_hi = i[-1]
    j_lo = j[0]
    j_hi = j[-1]
    assert i_lo > 0 and j_lo > 0
    assert i_hi < density.shape[0] - 1 and j_hi < density.shape[1] - 1

    # Compute the fixed solid pixels.
    fixed_solid = onp.zeros(design_shape, dtype=bool)
    fixed_solid[0, :] = density[i_lo - 1, j_lo : j_hi + 1] == 1
    fixed_solid[-1, :] = density[i_hi + 1, j_lo : j_hi + 1] == 1
    fixed_solid[:, 0] = density[i_lo : i_hi + 1, j_lo - 1] == 1
    fixed_solid[:, -1] = density[i_lo : i_hi + 1, j_hi + 1] == 1

    # Compute the fixed void pixels.
    fixed_void = onp.zeros(design_shape, dtype=bool)
    fixed_void[0, :] = density[i_lo - 1, j_lo : j_hi + 1] == 0
    fixed_void[-1, :] = density[i_hi + 1, j_lo : j_hi + 1] == 0
    fixed_void[:, 0] = density[i_lo : i_hi + 1, j_lo - 1] == 0
    fixed_void[:, -1] = density[i_lo : i_hi + 1, j_hi + 1] == 0

    return jnp.asarray(fixed_solid), jnp.asarray(fixed_void)


# -----------------------------------------------------------------------------
# Base class for ceviche challenges.
# -----------------------------------------------------------------------------


@dataclasses.dataclass
class CevicheChallenge(base.Challenge):
    """Defines a general ceviche challenge.

    The objective of the ceviche challenge is to find a component that whose
    transmission into its various ports lies within the target window defined
    by the transmission lower and upper bounds.

    Attributes:
        component: The component to be designed.
        transmission_lower_bound: The lower bound of the transmission window.
        transmission_upper_bound: The upper bound of the transmission window.
    """

    component: CevicheComponent
    transmission_lower_bound: jnp.ndarray
    transmission_upper_bound: jnp.ndarray

    def loss(self, response: CevicheResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        # Power transmission is the squared magnitude of the scattering parameters.
        transmission = jnp.abs(response.s_parameters) ** 2
        expected_shape = (
            len(response.wavelengths_nm),
            len(response.excite_port_idxs),
            len(self.component.ceviche_model.ports),
        )
        assert transmission.shape == expected_shape
        # Repeat the bounds to match the transmission shape. Each value in the bounds
        # is then interpreted as the bounds for a wavelength band.
        lb = _wavelength_bound(self.transmission_lower_bound, transmission.shape)
        ub = _wavelength_bound(self.transmission_upper_bound, transmission.shape)
        loss = transmission_loss.orthotope_smooth_transmission_loss(
            transmission=transmission,
            window_lower_bound=lb,
            window_upper_bound=ub,
            transmission_exponent=jnp.asarray(TRANSMISSION_EXPONENT),
            scalar_exponent=jnp.asarray(SCALAR_EXPONENT),
            axis=(1, 2),  # Generate a per-wavelength loss
        )
        return jnp.mean(loss)

    def eval_metric(self, response: CevicheResponse) -> jnp.ndarray:
        """Compute eval metric from the component `response`.

        The evaluation metric is computed from a psuedodistance to the target volume
        of the response space defined by the tranmission upper and lower bounds. A
        psuedodistance is defined for each element in the transmission array, and is
        a signed quantity with magnitude equal to the distance between a transmission
        value and its target, with sign being positive when the transmission is outside
        the target window, and negative inside.

        The distance is scaled by the size of the target window, and the negative
        maximum distance is taken as the eval metric.

        Positive values indicate that the target specification has been achieved.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        transmission = jnp.abs(response.s_parameters) ** 2
        lb = _wavelength_bound(self.transmission_lower_bound, transmission.shape)
        ub = _wavelength_bound(self.transmission_upper_bound, transmission.shape)
        psuedodistance = transmission_loss.elementwise_signed_psuedodistance_to_window(
            transmission,
            window_lower_bound=lb,
            window_upper_bound=ub,
        )
        scaled_psuedodistance = psuedodistance / (ub - lb)
        return -jnp.amax(scaled_psuedodistance)

    def metrics(
        self,
        response: CevicheResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        metrics = super().metrics(response, params, aux)
        return metrics


def _wavelength_bound(
    band_bound: jnp.ndarray, transmission_shape: Tuple[int, ...]
) -> jnp.ndarray:
    """Obtain per-wavelength bound compatible with `transmission_shape`."""
    assert len(transmission_shape) == 3
    assert band_bound.ndim <= 3

    # Add leading batch dimensions as needed.
    dims_to_add = range(0, len(transmission_shape) - band_bound.ndim)
    band_bound = jnp.expand_dims(band_bound, axis=dims_to_add)

    if not (transmission_shape[0] % band_bound.shape[0]) == 0:
        raise ValueError(
            f"Could not repeat `band_bound` with shape {band_bound.shape} to match "
            f"`transmission_shape` of {transmission_shape}; leading dimension size "
            f"must evenly divide the transmission shape."
        )

    repeats = transmission_shape[0] // band_bound.shape[0]
    repeated: jnp.ndarray = jnp.repeat(band_bound, repeats, axis=0)
    return repeated


# -----------------------------------------------------------------------------
# Constructors for ceviche challenges, including lightweight versions.
# -----------------------------------------------------------------------------


def beam_splitter(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Beamsplitter with 3.2 x 2.0 um design region.

    By default, standard simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.beam_splitter.model.BeamSplitterModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.BEAM_SPLITTER_SPEC,
            ),
            symmetries=defaults.BEAM_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.BEAM_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.BEAM_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_beam_splitter(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.LIGHTWEIGHT_SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.LIGHTWEIGHT_WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Beamsplitter with 3.2 x 2.0 um design and lightweight simulation params.

    By default, lightweight simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.beam_splitter.model.BeamSplitterModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.BEAM_SPLITTER_SPEC,
            ),
            symmetries=defaults.BEAM_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.BEAM_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.BEAM_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )


def mode_converter(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Mode converter with 1.6 x 1.6 um design and standard simulation params.

    By default, standard simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.mode_converter.model.ModeConverterModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.MODE_CONVERTER_SPEC,
            ),
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.MODE_CONVERTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.MODE_CONVERTER_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_mode_converter(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.LIGHTWEIGHT_SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.LIGHTWEIGHT_WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Mode converter with 1.6 x 1.6 um design and lightweight simulation params.

    By default, lightweight simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.mode_converter.model.ModeConverterModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.MODE_CONVERTER_SPEC,
            ),
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.MODE_CONVERTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.MODE_CONVERTER_TRANSMISSION_UPPER_BOUND,
    )


def waveguide_bend(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Waveguide bend with 1.6 x 1.6 um design and standard simulation params.

    By default, standard simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.waveguide_bend.model.WaveguideBendModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.WAVEGUIDE_BEND_SPEC,
            ),
            symmetries=defaults.WAVEGUIDE_BEND_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_waveguide_bend(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.LIGHTWEIGHT_SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.LIGHTWEIGHT_WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Waveguide bend with 1.6 x 1.6 um design and lightweight simulation params.

    By default, lightweight simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=cc.waveguide_bend.model.WaveguideBendModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.WAVEGUIDE_BEND_SPEC,
            ),
            symmetries=defaults.WAVEGUIDE_BEND_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WAVEGUIDE_BEND_TRANSMISSION_UPPER_BOUND,
    )


def wdm(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Demultiplexer with 6.4 x 6.4 um design and standard simulation params.

    By default, standard simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=wdm_model.WdmModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.wdm_spec(
                    design_extent_ij=u.Array([6400, 6400], u.nm),
                    design_wg_offset=u.Quantity(320, u.nm),
                    intended_sim_resolution=grid_spacing_nm * u.nm,
                ),
            ),
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WDM_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WDM_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_wdm(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.LIGHTWEIGHT_SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.LIGHTWEIGHT_WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Demultiplexer with 3.2 x 3.2 um design and lightweight simulation params.

    By default, lightweight simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=wdm_model.WdmModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.wdm_spec(
                    design_extent_ij=u.Array([3200, 3200], u.nm),
                    design_wg_offset=u.Quantity(320, u.nm),
                    intended_sim_resolution=grid_spacing_nm * u.nm,
                ),
            ),
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.WDM_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.WDM_TRANSMISSION_UPPER_BOUND,
    )


def power_splitter(
    minimum_width: int = defaults.MINIMUM_WIDTH,
    minimum_spacing: int = defaults.MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Power splitter with 1.6 x 1.6 um design and standard simulation params.

    By default, standard simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=wdm_model.WdmModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.wdm_spec(
                    design_extent_ij=u.Array([1600, 1600], u.nm),
                    design_wg_offset=u.Quantity(120, u.nm),
                    intended_sim_resolution=grid_spacing_nm * u.nm,
                ),
            ),
            symmetries=defaults.POWER_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.POWER_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.POWER_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )


def lightweight_power_splitter(
    minimum_width: int = defaults.LIGHTWEIGHT_MINIMUM_WIDTH,
    minimum_spacing: int = defaults.LIGHTWEIGHT_MINIMUM_SPACING,
    grid_spacing_nm: int = defaults.LIGHTWEIGHT_SIM_GRID_SPACING_NM,
    wavelengths_nm: Sequence[float] = defaults.LIGHTWEIGHT_WAVELENGTHS_NM,
    density_initializer: base.DensityInitializer = density_initializer,
) -> CevicheChallenge:
    """Power splitter with 1.6 x 1.6 um design and lightweight simulation params.

    By default, lightweight simulation parameters are used, but these may be overridden.

    Args:
        minimum_width: The minimum width target for the challenge, in pixels. The
            physical minimum width is approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        grid_spacing_nm: The spacing of the simulation and design grid.
        wavelengths_nm: The wavelengths for which the response is computed.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.

    Returns:
        The configured `CevicheChallenge`.
    """
    return CevicheChallenge(
        component=CevicheComponent(
            ceviche_model=wdm_model.WdmModel(
                params=cc.params.CevicheSimParams(
                    resolution=grid_spacing_nm * u.nm,
                    wavelengths=u.Array(wavelengths_nm, u.nm),
                ),
                spec=defaults.wdm_spec(
                    design_extent_ij=u.Array([1600, 1600], u.nm),
                    design_wg_offset=u.Quantity(120, u.nm),
                    intended_sim_resolution=grid_spacing_nm * u.nm,
                ),
            ),
            symmetries=defaults.POWER_SPLITTER_SYMMETRIES,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            density_initializer=density_initializer,
        ),
        transmission_lower_bound=defaults.POWER_SPLITTER_TRANSMISSION_LOWER_BOUND,
        transmission_upper_bound=defaults.POWER_SPLITTER_TRANSMISSION_UPPER_BOUND,
    )



"""Defines the photon extractor challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Tuple

from fmmax import basis, fmm  # type: ignore[import-untyped]
from jax import numpy as jnp
from totypes import symmetry, types

from invrs_gym.challenges import base
from invrs_gym.challenges.extractor import component as extractor_component
from invrs_gym.utils import initializers

ENHANCEMENT_FLUX_PER_DIPOLE = "enhancement_flux_per_dipole"
ENHANCEMENT_FLUX_TOTAL = "enhancement_flux_total"
ENHANCEMENT_DOS_PER_DIPOLE = "enhancement_dos_per_dipole"
ENHANCEMENT_DOS_TOTAL = "enhancement_dos_total"


density_initializer = functools.partial(
    initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class PhotonExtractorChallenge(base.Challenge):
    """Defines the photon extractor challenge.

    The challenge is based on "Inverse-designed photon extractors for optically
    addressable defect qubits" by Chakravarthi et al. It involves optimizing a GaP
    patterned layer on diamond substrate above an implanted nitrogen vacancy defect.
    An oxide hard mask used to pattern the GaP is left in place after the etch.

    The goal of the optimization is to maximize extraction of 637 nm emission, i.e.
    to maximize the power coupled from the defect to the ambient above the extractor.

    https://opg.optica.org/optica/fulltext.cfm?uri=optica-7-12-1805

    Attributes:
        component: The component to be designed.
    """

    component: extractor_component.ExtractorComponent

    def loss(self, response: extractor_component.ExtractorResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        # The response should have a length-3 trailing axis, corresponding to x, y,
        # and z-oriented dipoles.
        assert response.collected_power.shape[-1] == 3
        return -jnp.mean(response.collected_power)

    def eval_metric(
        self,
        response: extractor_component.ExtractorResponse,
    ) -> jnp.ndarray:
        """Computes the eval metric from the component `response`.

        The evaluation metric is the enhancement in collected power, summing over all
        wavelengths and dipole polarizations.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        assert response.collected_power.shape[-1] == 3
        enhancement_flux_total = jnp.sum(response.collected_power) / jnp.sum(
            response.bare_substrate_collected_power
        )
        return enhancement_flux_total

    def metrics(
        self,
        response: extractor_component.ExtractorResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Args:
            response: The response of the extractor component.
            params: The parameters where the response was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary, with the following quantities:
                - the per-dipole flux enhancement
                - the total flux enhancment
                - the per-dipole density of states enhancement
                - the total density of states enhancement
        """
        metrics = super().metrics(response, params, aux)
        enhancement_flux_per_dipole = (
            response.collected_power / response.bare_substrate_collected_power
        )
        enhancement_flux_total = jnp.sum(response.collected_power) / jnp.sum(
            response.bare_substrate_collected_power
        )
        enhancement_dos_per_dipole = (
            response.emitted_power / response.bare_substrate_emitted_power
        )
        enhancement_dos_total = jnp.sum(response.emitted_power) / jnp.sum(
            response.bare_substrate_emitted_power
        )
        metrics.update(
            {
                ENHANCEMENT_FLUX_PER_DIPOLE: enhancement_flux_per_dipole,
                ENHANCEMENT_FLUX_TOTAL: enhancement_flux_total,
                ENHANCEMENT_DOS_PER_DIPOLE: enhancement_dos_per_dipole,
                ENHANCEMENT_DOS_TOTAL: enhancement_dos_total,
            }
        )
        return metrics


EXTRACTOR_SPEC = extractor_component.ExtractorSpec(
    permittivity_ambient=(1.0 + 0.0j) ** 2,
    permittivity_oxide=(1.46 + 0.0j) ** 2,
    permittivity_extractor=(3.31 + 0.0j) ** 2,
    permittivity_substrate=(2.4102 + 0.0j) ** 2,
    thickness_ambient=1.0,
    thickness_oxide=0.13,
    thickness_extractor=0.25,
    thickness_substrate_before_source=0.1,
    thickness_substrate_after_source=0.9,
    width_design_region=1.5,
    width_padding=0.25,
    width_pml=0.4,
    fwhm_source=0.05,
    offset_monitor_source=0.025,
    offset_monitor_ambient=0.4,
    width_monitor_ambient=1.5,
    grid_spacing=0.005,
)

EXTRACTOR_SIM_PARAMS = extractor_component.ExtractorSimParams(
    wavelength=0.637,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    approximate_num_terms=1200,
    truncation=basis.Truncation.CIRCULAR,
)

SYMMETRIES: Tuple[str, ...] = (
    symmetry.REFLECTION_N_S,
    symmetry.REFLECTION_E_W,
    symmetry.REFLECTION_NE_SW,
    symmetry.REFLECTION_NW_SE,
)

# Minimum width and spacing are 50 nm for the default dimensions.
MINIMUM_WIDTH = 10
MINIMUM_SPACING = 10


def photon_extractor(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    density_initializer: base.DensityInitializer = density_initializer,
    spec: extractor_component.ExtractorSpec = EXTRACTOR_SPEC,
    sim_params: extractor_component.ExtractorSimParams = EXTRACTOR_SIM_PARAMS,
    symmetries: Tuple[str, ...] = SYMMETRIES,
) -> PhotonExtractorChallenge:
    """Photon extractor with 1.5 x 1.5 um design region.

    The challenge is based on "Inverse-designed photon extractors for optically
    addressable defect qubits" by Chakravarthi et al. It involves optimizing a GaP
    patterned layer on diamond substrate above an implanted nitrogen vacancy defect.
    An oxide hard mask used to pattern the GaP is left in place after the etch.

    The goal of the optimization is to maximize extraction of 637 nm emission, i.e.
    to maximize the power coupled from the defect to the ambient above the extractor.

    https://opg.optica.org/optica/fulltext.cfm?uri=optica-7-12-1805

    Args:
        minimum_width: The minimum width target for the challenge, in pixels.  The
            default value of 10 corresponds to a physical size of approximately 50 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        spec: Defines the physical specification of the photon extractor.
        sim_params: Defines the simulation settings of the photon extractor.
        symmetries: Defines the symmetries of the photon extractor.

    Returns:
        The `PhotonExtractorChallenge`.
    """
    return PhotonExtractorChallenge(
        component=extractor_component.ExtractorComponent(
            spec=spec,
            sim_params=sim_params,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
    )



"""Defines the bayer color sorter component.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Any, Callable, Dict, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as onp
from fmmax import basis, fields, fmm, scattering  # type: ignore[import-untyped]
from invrs_gym import utils
from invrs_gym.challenges import base
from jax import tree_util
from totypes import json_utils, types

from invrs_gym.utils import materials

NDArray = onp.ndarray[Any, Any]


Params = Dict[str, types.BoundedArray | types.Density2DArray]
ThicknessInitializer = Callable[[jax.Array, types.BoundedArray], types.BoundedArray]


THICKNESS_METASURFACE = "thickness_metasurface"
DENSITY_METASURFACE = "density_metasurface"
OFFSET_MONITOR_SUBSTRATE = "offset_monitor_substrate"

EFIELD_XY = "efield_xy"
HFIELD_XY = "hfield_xy"
POYNTING_FLUX_XY = "poynting_flux_xy"
COORDINATES_XY = "coordinates_xy"
TRANSMITTED_POWER = "transmitted_power"

EFIELD_XZ = "efield_xz"
HFIELD_XZ = "hfield_xz"
POYNTING_FLUX_XZ = "poynting_flux_xz"
COORDINATES_XZ = "coordinates_xz"

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0


@dataclasses.dataclass
class BayerSpec:
    """Defines the physical specification of a bayer color sorter.

    Attributes:
        material_ambient: The ambient material.
        material_metasurface_solid: The material comprising metasurface solid regions.
        material_metasurface_void: The material comprising metasurface void regions.
        material_substrate: The substrate material.
        thickness_ambient: The thickness of the ambient.
        thickness_metasurface: Thicknesses of the metasurface layers.
        thickness_substrate: The thickness of the substrate.
        offset_monitor_substrate: Offset of the monitor plane from the interface
            between spacer and substrate.
        pixel_size: The size of a pixel in the bayer array.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    material_ambient: str
    material_metasurface_solid: str
    material_metasurface_void: str
    material_substrate: str

    thickness_ambient: float
    thickness_metasurface: types.BoundedArray
    thickness_substrate: float
    offset_monitor_substrate: types.BoundedArray

    pixel_size: float
    grid_spacing: float

    @property
    def period_x(self) -> float:
        """Returns the size of the unit cell in the `x` direction."""
        return 2 * self.pixel_size

    @property
    def period_y(self) -> float:
        """Returns the size of the unit cell in the `y` direction."""
        return 2 * self.pixel_size

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        return (
            int(jnp.ceil(self.period_x / self.grid_spacing)),
            int(jnp.ceil(self.period_y / self.grid_spacing)),
        )


@dataclasses.dataclass
class BayerSimParams:
    """Parameters that configure the simulation of a bayer color sorter.

    Attributes:
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    wavelength: float | jnp.ndarray
    polar_angle: float | jnp.ndarray
    azimuthal_angle: float | jnp.ndarray
    formulation: fmm.Formulation
    approximate_num_terms: int
    truncation: basis.Truncation


@dataclasses.dataclass
class BayerResponse:
    """Contains the response of the bayer color sorter.

    Attributes:
        wavelength: The wavelength for the sorter response.
        polar_angle: The polar angle for the sorter response.
        azimuthal_angle: The azimuthal angle for the sorter response.
        transmission: The transmission into the four pixels for each polarization.
        reflection: The reflection back to the ambient for the each polarizations.
    """

    wavelength: jnp.ndarray
    polar_angle: jnp.ndarray
    azimuthal_angle: jnp.ndarray
    transmission: jnp.ndarray
    reflection: jnp.ndarray


json_utils.register_custom_type(BayerResponse)

tree_util.register_pytree_node(
    BayerResponse,
    lambda r: (
        (
            r.wavelength,
            r.polar_angle,
            r.azimuthal_angle,
            r.transmission,
            r.reflection,
        ),
        None,
    ),
    lambda _, children: BayerResponse(*children),
)


class BayerComponent(base.Component):
    """Defines a bayer color sorter component."""

    def __init__(
        self,
        spec: BayerSpec,
        sim_params: BayerSimParams,
        thickness_initializer: ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the bayer color sorter component.

        Args:
            spec: Defines the physical specification of the bayer color sorter.
            sim_params: Defines simulation parameters for the sorter.
            thickness_initializer: Callable which returns the initial thickness for
                a layer from a random key and a bounded array with value equal the
                thickness from `spec`.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_density = _seed_density(
            grid_shape=self.spec.grid_shape, **seed_density_kwargs
        )
        self.expansion = basis.generate_expansion(
            primitive_lattice_vectors=basis.LatticeVectors(
                u=self.spec.period_x * basis.X,
                v=self.spec.period_y * basis.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for the bayer color sorter component."""
        key_thickness, key_density, key_offset = jax.random.split(key, num=3)
        params = {
            THICKNESS_METASURFACE: self.thickness_initializer(
                key_thickness, self.spec.thickness_metasurface
            ),
            DENSITY_METASURFACE: self.density_initializer(
                key_density, self.seed_density
            ),
            OFFSET_MONITOR_SUBSTRATE: self.thickness_initializer(
                key_offset, self.spec.offset_monitor_substrate
            ),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        *,
        wavelength: Optional[float | jnp.ndarray] = None,
        polar_angle: Optional[float | jnp.ndarray] = None,
        azimuthal_angle: Optional[float | jnp.ndarray] = None,
        expansion: Optional[basis.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[BayerResponse, base.AuxDict]:
        """Computes the response of the sorter.

        Args:
            params: The parameters defining the sorter, with structure matching that
                of the parameters returned by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            polar_angle: Optional polar angle to override the default.
            azimuthal_angle: Optional azimuthal angle to override the default.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, computes fields in an xz cross section.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        if polar_angle is None:
            polar_angle = self.sim_params.polar_angle
        if azimuthal_angle is None:
            azimuthal_angle = self.sim_params.azimuthal_angle

        spec = dataclasses.replace(
            self.spec,
            thickness_metasurface=(
                params[THICKNESS_METASURFACE]  # type: ignore[arg-type]
            ),
            offset_monitor_substrate=(
                params[OFFSET_MONITOR_SUBSTRATE]  # type: ignore[arg-type]
            ),
        )
        return simulate_color_sorter(
            density=params[DENSITY_METASURFACE],  # type: ignore[arg-type]
            spec=spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(polar_angle),
            azimuthal_angle=jnp.asarray(azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )


def _seed_density(grid_shape: Tuple[int, int], **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a bayer sorter component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a sorter component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(True, True),
        **kwargs,
    )


def simulate_color_sorter(
    density: types.Density2DArray,
    spec: BayerSpec,
    wavelength: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
    compute_fields: bool,
) -> Tuple[BayerResponse, base.AuxDict]:
    """Simulates a bayer color sorter.

    Args:
        density: Defines the pattern of the metasurface layer.
        spec: Defines the physical specification of the sorter.
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.
        compute_fields: If `True`, fields in an xz cross section are computed and
            and included in the `aux` return variable.

    Returns:
        The `BayerResponse`, and a dictionary containing the auxilliary quantities.
    """
    primitive_lattice_vectors = basis.LatticeVectors(
        u=spec.period_x * basis.X,
        v=spec.period_y * basis.Y,
    )
    in_plane_wavevector = basis.plane_wave_in_plane_wavevector(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        permittivity=materials.permittivity(spec.material_ambient, wavelength),
    )

    # -------------------------------------------------------------------------
    # Layer permittivities, thicknesses, and eigensolve results.
    # -------------------------------------------------------------------------

    def permittivity_fn(material_name):
        shape = (wavelength.size, 1, 1)
        return materials.permittivity(material_name, wavelength).reshape(shape)

    eigensolve_fn = functools.partial(
        fmm.eigensolve_isotropic_media,
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        formulation=formulation,
    )

    # Eigensolve for the uniform ambient and substrate layers.
    solve_result_ambient = eigensolve_fn(
        permittivity=permittivity_fn(spec.material_ambient)
    )
    solve_result_substrate = eigensolve_fn(
        permittivity=permittivity_fn(spec.material_substrate)
    )

    # Eigensolve for the patterned metasurface layer.
    density_array = utils.transforms.rescaled_density_array(
        density=density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    permittivity_metasurface = utils.transforms.interpolate_permittivity(
        permittivity_solid=permittivity_fn(spec.material_metasurface_solid),
        permittivity_void=permittivity_fn(spec.material_metasurface_void),
        density=density_array,
    )
    solve_result_metasurface = eigensolve_fn(permittivity=permittivity_metasurface)

    # Assemble the list of all layer solve results.
    layer_solve_results = [
        solve_result_ambient,
        solve_result_metasurface,
        solve_result_substrate,
    ]
    layer_thicknesses = [
        jnp.asarray(spec.thickness_ambient),
        jnp.asarray(spec.thickness_metasurface.array),
        jnp.asarray(spec.thickness_substrate),
    ]

    # -------------------------------------------------------------------------
    # Scattering matrix assembly.
    # -------------------------------------------------------------------------

    if compute_fields:
        # If the field calculation is desired, compute the interior scattering
        # matrices. For each layer in the stack, the interior scattering matrices
        # consist of a pair of matrices, one for the substack below the layer, and
        # one for the substack above the layer.
        s_matrices_interior = scattering.stack_s_matrices_interior(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )
        s_matrix = s_matrices_interior[-1][0]
    else:
        s_matrix = scattering.stack_s_matrix(layer_solve_results, layer_thicknesses)

    # -------------------------------------------------------------------------
    # Excitation and wave amplitude calculation.
    # -------------------------------------------------------------------------

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate wave amplitudes for forward-going waves with both polarizations.
    fwd_ambient_start = jnp.zeros((2 * n, 2), dtype=complex)
    fwd_ambient_start = fwd_ambient_start.at[0, 0].set(1)
    fwd_ambient_start = fwd_ambient_start.at[n, 1].set(1)

    # Compute the backward-goingpower at the start of the ambient.
    bwd_ambient_end = s_matrix.s21 @ fwd_ambient_start
    bwd_ambient_start = fields.propagate_amplitude(
        bwd_ambient_end,
        jnp.asarray(spec.thickness_ambient),
        layer_solve_result=layer_solve_results[0],
    )
    sz_fwd_ambient, sz_bwd_ambient = fields.amplitude_poynting_flux(
        forward_amplitude=fwd_ambient_start,
        backward_amplitude=bwd_ambient_start,
        layer_solve_result=layer_solve_results[0],
    )
    sz_fwd_ambient_sum = jnp.sum(jnp.abs(sz_fwd_ambient), axis=-2)
    sz_bwd_ambient_sum = jnp.sum(jnp.abs(sz_bwd_ambient), axis=-2)
    reflection = jnp.abs(sz_bwd_ambient_sum) / jnp.abs(sz_fwd_ambient_sum)

    # Compute the forward-going and backward-going wave amplitudes in the substrate,
    # a distance `spec.offset_monitor_substrate` from the start of the substrate.
    fwd_substrate_start = s_matrix.s11 @ fwd_ambient_start
    fwd_substrate_monitor, bwd_substrate_monitor = fields.colocate_amplitudes(
        fwd_substrate_start,
        jnp.zeros_like(fwd_substrate_start),
        z_offset=jnp.asarray(spec.offset_monitor_substrate.array),
        layer_solve_result=layer_solve_results[-1],
        layer_thickness=layer_thicknesses[-1],
    )

    # -------------------------------------------------------------------------
    # Compute fields at the monitor plane.
    # -------------------------------------------------------------------------

    # Compute transmitted power directly from the wave amplitudes.
    sz_fwd_monitor, _ = fields.amplitude_poynting_flux(
        forward_amplitude=fwd_substrate_monitor,
        backward_amplitude=bwd_substrate_monitor,
        layer_solve_result=layer_solve_results[-1],
    )
    sz_fwd_monitor_sum = jnp.sum(jnp.abs(sz_fwd_monitor), axis=-2)
    transmission = jnp.abs(sz_fwd_monitor_sum) / jnp.abs(sz_fwd_ambient_sum)

    # Compute electric and magnetic fields at the monitor plane in their Fourier
    # representation, and then on the real-space grid.
    ef, hf = fields.fields_from_wave_amplitudes(
        forward_amplitude=fwd_substrate_monitor,
        backward_amplitude=bwd_substrate_monitor,
        layer_solve_result=layer_solve_results[-1],
    )
    grid_shape: Tuple[int, int]
    grid_shape = density.shape[-2:]  # type: ignore[assignment]
    (ex, ey, ez), (hx, hy, hz), (x, y) = fields.fields_on_grid(
        electric_field=ef,
        magnetic_field=hf,
        layer_solve_result=layer_solve_results[-1],
        shape=grid_shape,
        num_unit_cells=(1, 1),
    )
    batch_shape = layer_solve_results[0].batch_shape
    assert ex.shape == batch_shape + grid_shape + (2,)

    # Compute the Poynting flux on the real-space grid at the monitor plane.
    sz = _time_average_z_poynting_flux((ex, ey, ez), (hx, hy, hz))
    assert sz.shape == batch_shape + grid_shape + (2,)

    # Create masks selecting the four pixelss.
    pixel_mask = _pixel_mask(grid_shape=grid_shape)
    assert pixel_mask.shape == grid_shape + (1, 4)

    # Use the mask to compute the time average Poynting flux into each quadrant. The
    # trailing two dimensions have shape `(2, 4)`; index `(i, j)` corresponds
    # to power for the `i` excitation (i.e. polarization) in the `j` pixel.
    pixel_transmission = jnp.mean(pixel_mask * sz[..., jnp.newaxis], axis=(-4, -3))
    pixel_transmission /= sz_fwd_ambient_sum[..., jnp.newaxis]

    assert pixel_transmission.shape == batch_shape + (2, 4)

    response = BayerResponse(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        transmission=pixel_transmission,
        reflection=reflection,
    )

    aux = {
        EFIELD_XY: (ex, ey, ez),
        HFIELD_XY: (hx, hy, hz),
        POYNTING_FLUX_XY: sz,
        COORDINATES_XY: (x, y),
        TRANSMITTED_POWER: transmission,
    }

    # -------------------------------------------------------------------------
    # Optionally compute fields in an xz cross section.
    # -------------------------------------------------------------------------

    if compute_fields:
        amplitudes_interior = fields.stack_amplitudes_interior(
            s_matrices_interior=s_matrices_interior,
            forward_amplitude_0_start=fwd_ambient_start,
            backward_amplitude_N_end=jnp.zeros_like(fwd_ambient_start),
        )
        layer_znum = tuple(
            [int(jnp.round(t / spec.grid_spacing) + 1) for t in layer_thicknesses]
        )
        x = jnp.stack(
            [jnp.arange(0, spec.grid_shape[0]) * spec.grid_spacing] * 2,
            axis=-1,
        )
        y = jnp.stack(
            [
                jnp.ones((spec.grid_shape[0],)) * spec.pixel_size * 0.5,
                jnp.ones((spec.grid_shape[0],)) * spec.pixel_size * 1.5,
            ],
            axis=-1,
        )
        assert x.shape == y.shape == (spec.grid_shape[0], 2)

        (
            (ex, ey, ez),
            (hx, hy, hz),
            (xf, yf, zf),
        ) = fields.stack_fields_3d_on_coordinates(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
            layer_znum=layer_znum,
            x=x,
            y=y,
        )
        aux.update(
            {
                EFIELD_XZ: (ex, ey, ez),
                HFIELD_XZ: (hx, hy, hz),
                COORDINATES_XZ: (xf, yf, zf),
            }
        )

    return response, aux


# -----------------------------------------------------------------------------
# Functions related to the densities of various layers in the stack.
# -----------------------------------------------------------------------------


def _pixel_mask(grid_shape: Tuple[int, int]) -> jnp.ndarray:
    """Return masks that select the four quadrants of a sorter.

    The quadrants are numbered as follows:     0  |  1
                                             -----------
                                               2  |  3

    Args:
        grid_shape: The shape of the grid for which to return the mask.

    Returns:
        The quadrant mask, with shape `grid_shape + (1, 4)`.
    """
    quadrant_mask = jnp.zeros(grid_shape + (1, 4))
    xdim = grid_shape[0] // 2
    ydim = grid_shape[1] // 2
    quadrant_mask = quadrant_mask.at[:xdim, :ydim, 0, 0].set(1)  # nw
    quadrant_mask = quadrant_mask.at[:xdim, ydim:, 0, 1].set(1)  # ne
    quadrant_mask = quadrant_mask.at[xdim:, :ydim, 0, 2].set(1)  # sw
    quadrant_mask = quadrant_mask.at[xdim:, ydim:, 0, 3].set(1)  # se
    return quadrant_mask


def _time_average_z_poynting_flux(
    electric_fields: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_fields: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """Computes the time-average z-directed Poynting flux, given the physical fields."""
    ex, ey, _ = electric_fields
    hx, hy, _ = magnetic_fields
    return jnp.real(ex * jnp.conj(hy) - ey * jnp.conj(hx))



"""Defines the meta-atom library challenge.

Copyright (c) 2024 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Sequence, Tuple

import jax
import jax.numpy as jnp
from fmmax import basis, fmm
from totypes import types

from invrs_gym.challenges import base
from invrs_gym.challenges.library import component as library_component
from invrs_gym.utils import initializers, materials

METAGRATING_EFFICIENCY_RHCP = "metagrating_efficiency_rhcp"
METAGRATING_EFFICIENCY_LHCP = "metagrating_efficiency_lhcp"
METAGRATING_RELATIVE_EFFICIENCY_RHCP = "metagrating_relative_efficiency_rhcp"
METAGRATING_RELATIVE_EFFICIENCY_LHCP = "metagrating_relative_efficiency_lhcp"
METAGRATING_AVERAGE_EFFICIENCY = "metagrating_average_efficiency"
METAGRATING_AVERAGE_RELATIVE_EFFICIENCY = "metagrating_average_relative_efficiency"


@dataclasses.dataclass
class LibraryChallenge(base.Challenge):
    """The meta-atom library design challenge."""

    component: library_component.LibraryComponent

    def loss(self, response: library_component.LibraryResponse) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`."""
        (efficiency_rhcp, efficiency_lhcp), _ = _metagrating_efficiency(
            response, self.component.spec
        )
        loss_rhcp = jnp.sum(jnp.abs(1 - efficiency_rhcp)) ** 2
        loss_lhcp = jnp.sum(jnp.abs(1 - efficiency_lhcp)) ** 2
        return loss_rhcp + loss_lhcp

    def eval_metric(
        self,
        response: library_component.LibraryResponse,
    ) -> jnp.ndarray:
        """Computes the eval metric from the component `response`.

        The eval metric considers a metagrating assembled from the eight meta-atoms in
        the library. The relative efficiency of the grating (i.e. transmitted power
        into the target order divided by total transmitted power) is computed for each
        wavelength and the two incident polarization states. The eval metric is the
        minimum relative efficiency among all these cases.

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        (
            _,
            (relative_efficiency_rhcp, relative_efficiency_lhcp),
        ) = _metagrating_efficiency(response, self.component.spec)
        return jnp.minimum(
            jnp.amin(relative_efficiency_rhcp),
            jnp.amin(relative_efficiency_lhcp),
        )

    def metrics(
        self,
        response: library_component.LibraryResponse,
        params: library_component.Params,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics.

        Several challenge metrics relate to the performance of a metagrating assembled
        from the meta-atom library.

        Args:
            response: The response of the meta-atom library.
            params: The parameters where the response was evaluated.
            aux: The auxilliary quantities returned by the component response method.

        Returns:
            The metrics dictionary, with the following quantities:
                - Per-wavelength, per-polarization metagrating efficiency for RHCP
                  excitation.
                - Per-wavelength, per-polarization metagrating efficiency for LHCP
                  excitation.
                - Per-wavelength, per-polarization metagrating relative efficiency for
                  RHCP excitation.
                - Per-wavelength, per-polarization metagrating relative efficiency for
                  LHCP excitation.
                - Average metagrating efficiency.
                - Average metagrating relative efficiency.
        """
        metrics = super().metrics(response, params, aux)
        (
            (efficiency_rhcp, efficiency_lhcp),
            (relative_efficiency_rhcp, relative_efficiency_lhcp),
        ) = _metagrating_efficiency(response, self.component.spec)
        metrics.update(
            {
                METAGRATING_EFFICIENCY_RHCP: efficiency_rhcp,
                METAGRATING_EFFICIENCY_LHCP: efficiency_lhcp,
                METAGRATING_RELATIVE_EFFICIENCY_RHCP: relative_efficiency_rhcp,
                METAGRATING_RELATIVE_EFFICIENCY_LHCP: relative_efficiency_lhcp,
                METAGRATING_AVERAGE_EFFICIENCY: jnp.mean(
                    jnp.asarray([efficiency_rhcp, efficiency_lhcp])
                ),
                METAGRATING_AVERAGE_RELATIVE_EFFICIENCY: jnp.mean(
                    jnp.asarray([relative_efficiency_rhcp, relative_efficiency_lhcp])
                ),
            }
        )
        return metrics


def _metagrating_efficiency(
    response: library_component.LibraryResponse,
    spec: library_component.LibrarySpec,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray]]:
    """Return efficiency of a metagrating assembled from the meta-atom libary."""

    # Scale the transmission coefficients to account for the difference in
    # material permittivity. With this rescaling, perfect transmission will
    # result in transmission coefficients with a norm of 1.
    permittivity_substrate = materials.permittivity(
        spec.material_substrate, response.wavelength
    )
    permittivity_ambient = materials.permittivity(
        spec.material_ambient, response.wavelength
    )
    scalar = jnp.sqrt(
        jnp.sqrt(permittivity_substrate).real / jnp.sqrt(permittivity_ambient).real
    )
    transmission_rhcp_conserved = response.transmission_rhcp[..., 0] * scalar
    transmission_rhcp_converted = response.transmission_rhcp[..., 1] * scalar
    transmission_lhcp_conserved = response.transmission_lhcp[..., 1] * scalar
    transmission_lhcp_converted = response.transmission_lhcp[..., 0] * scalar

    _fft = functools.partial(jnp.fft.fft, axis=0, norm="forward")
    efficiency_per_order_rhcp = (
        jnp.abs(_fft(transmission_rhcp_conserved)) ** 2
        + jnp.abs(_fft(transmission_rhcp_converted)) ** 2
    )
    efficiency_per_order_lhcp = (
        jnp.abs(_fft(transmission_lhcp_conserved)) ** 2
        + jnp.abs(_fft(transmission_lhcp_converted)) ** 2
    )

    efficiency_rhcp = efficiency_per_order_rhcp[1, ...]
    efficiency_lhcp = efficiency_per_order_lhcp[1, ...]

    total_efficiency_rhcp = jnp.sum(efficiency_per_order_rhcp, axis=0)
    total_efficiency_lhcp = jnp.sum(efficiency_per_order_lhcp, axis=0)

    relative_efficiency_rhcp = efficiency_rhcp / total_efficiency_rhcp
    relative_efficiency_lhcp = efficiency_lhcp / total_efficiency_lhcp

    return (
        (efficiency_rhcp, efficiency_lhcp),
        (relative_efficiency_rhcp, relative_efficiency_lhcp),
    )


def library_density_initializer(
    key: jax.Array,
    seed_density: types.Density2DArray,
    relative_mean_range: float = 0.8,
    relative_noise_amplitude=0.1,
    resize_method: jax.image.ResizeMethod = jax.image.ResizeMethod.CUBIC,
) -> types.Density2DArray:
    """Returns the random initial density."""
    assert seed_density.ndim == 3
    library_size = seed_density.shape[0]
    keys = jax.random.split(key, num=library_size)
    relative_mean_values = (
        jnp.arange(library_size) + 0.5
    ) / library_size * relative_mean_range + (1 - relative_mean_range) / 2
    return jax.vmap(
        initializers.noisy_density_initializer,
        in_axes=(0, 0, 0, None, None),
    )(keys, seed_density, relative_mean_values, relative_noise_amplitude, resize_method)


MINIMUM_WIDTH = 12
MINIMUM_SPACING = 12
LIBRARY_SIZE = 8

LIBRARY_SPEC = library_component.LibrarySpec(
    material_ambient=materials.VACUUM,
    material_metasurface_solid=library_component.TIO2_CHEN,
    material_metasurface_void=materials.VACUUM,
    material_substrate=materials.SIO2,
    background_extinction_coeff=0.0001,
    thickness_ambient=1.2,
    thickness_metasurface=types.BoundedArray(
        array=0.6, lower_bound=0.5, upper_bound=0.7
    ),
    thickness_substrate=0.2,
    pitch=0.4,
    frame_width=0.03,
    grid_spacing=0.005,
)

LIBRARY_SIM_PARAMS = library_component.LibrarySimParams(
    wavelength=jnp.asarray([0.45, 0.55, 0.65]),
    approximate_num_terms=200,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    truncation=basis.Truncation.CIRCULAR,
)

SYMMETRIES = ("reflection_n_s", "reflection_e_w")


def meta_atom_library(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    library_size: int = LIBRARY_SIZE,
    spec: library_component.LibrarySpec = LIBRARY_SPEC,
    sim_params: library_component.LibrarySimParams = LIBRARY_SIM_PARAMS,
    thickness_initializer: base.ThicknessInitializer = (
        initializers.identity_initializer
    ),
    density_initializer: base.DensityInitializer = library_density_initializer,
    symmetries: Sequence[str] = SYMMETRIES,
) -> LibraryChallenge:
    """Return the meta-atom library design challenge.

    The library design challenge is based on "Dispersion-engineered metasurfaces
    reaching broadband 90% relative diffraction efficiency" by Chen et al.
    https://www.nature.com/articles/s41467-023-38185-2

    Args:
        minimum_width: The minimum width target for the challenge, in pixels.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        library_size: The number of meta-atoms in the library.
        spec: Defines the physical specification of the meta-atom library.
        sim_params: Defines the simulation parameters of the meta-atom library.
        thickness_initializer: Callable which returns the initial thickness, given a
            key and seed thickness.
        density_initializer: Callable which returns the initial density, given a
            key and seed density.
        symmetries: Defines the symmetries of the meta-atoms.

    Returns:
        The `LibraryChallenge`.
    """
    return LibraryChallenge(
        component=library_component.LibraryComponent(
            spec=spec,
            sim_params=sim_params,
            library_size=library_size,
            thickness_initializer=thickness_initializer,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        )
    )



"""Base types for challenges and components.

Copyright (c) 2023 The INVRS-IO authors.
"""

import abc
from typing import Any, Callable, Dict, Tuple

import fmmax
import jax
import jax.numpy as jnp
from totypes import json_utils, types

from invrs_gym.utils import metrics

AuxDict = Dict[str, Any]
PyTree = Any
DensityInitializer = Callable[[jax.Array, types.Density2DArray], types.Density2DArray]
ThicknessInitializer = Callable[[jax.Array, types.BoundedArray], types.BoundedArray]

BINARIZATION_DEGREE = "binarization_degree"


class Component(abc.ABC):
    """Base class for components.

    The component represents the physical structure to be optimized. Initial parameters
    are returned by the `init` method; these can include types having metadata that
    specifies bounds on parameter values or target characteristics, e.g. a minimum
    feature size.

    The component has a default "excitation", e.g. illumination with a particular
    wavelength of light from a specific direction; the response of the component to the
    excitation is obtained with the `response` method. The `response` method of classes
    inheriting from `Component` will generally have optional arguments to modify the
    excitation condition.
    """

    @abc.abstractmethod
    def init(self, key: jax.Array):
        """Returns the initial parameters given a `PRNGKey`."""

    @abc.abstractmethod
    def response(self, params: PyTree) -> Tuple[Any, AuxDict]:
        """Returns the response of the component for the given `params`."""


class Challenge(abc.ABC):
    """Base class for challenges.

    Challenges consist of a `Component` to be optimized and a loss function which
    returns a scalar loss given the component response.

    Challenges also include an `eval_metric` function whcih computes a scalar metric
    (independent of the loss), which can be used to evaluate the quality of solutions
    to the challenge. Higher `eval_metric` corresponds to a better solution.

    Challenges also include a `metrics` function which computes additional quantities
    that may be useful assessing the quality of a particular component.

    Attributes:
        component: The `Component` to be optimized.
    """

    component: Component

    @abc.abstractmethod
    def loss(self, response: Any) -> jnp.ndarray:
        """Compute scalar loss for the `response`."""

    @abc.abstractmethod
    def eval_metric(self, response: Any) -> jnp.ndarray:
        """Compute scalar evaluation metric for the `response`."""

    def metrics(self, response: Any, params: PyTree, aux: AuxDict) -> AuxDict:
        """Compute metrics for a component response and associated quantities."""
        del response, aux
        return {BINARIZATION_DEGREE: metrics.binarization_degree(params)}


# Several challenges use the `fmmax` simulator, and contain `fmmax` custom objects in
# their responses. Ensure these are serializable by registering with totypes.
json_utils.register_custom_type(fmmax.basis.Expansion)



from invrs_gym.challenges.bayer.challenge import bayer_sorter as bayer_sorter
from invrs_gym.challenges.ceviche.challenge import (
    beam_splitter as ceviche_beam_splitter,
)
from invrs_gym.challenges.ceviche.challenge import (
    lightweight_beam_splitter as ceviche_lightweight_beam_splitter,
)
from invrs_gym.challenges.ceviche.challenge import (
    lightweight_mode_converter as ceviche_lightweight_mode_converter,
)
from invrs_gym.challenges.ceviche.challenge import (
    lightweight_power_splitter as ceviche_lightweight_power_splitter,
)
from invrs_gym.challenges.ceviche.challenge import (
    lightweight_waveguide_bend as ceviche_lightweight_waveguide_bend,
)
from invrs_gym.challenges.ceviche.challenge import (
    lightweight_wdm as ceviche_lightweight_wdm,
)
from invrs_gym.challenges.ceviche.challenge import (
    mode_converter as ceviche_mode_converter,
)
from invrs_gym.challenges.ceviche.challenge import (
    power_splitter as ceviche_power_splitter,
)
from invrs_gym.challenges.ceviche.challenge import (
    waveguide_bend as ceviche_waveguide_bend,
)
from invrs_gym.challenges.ceviche.challenge import wdm as ceviche_wdm
from invrs_gym.challenges.diffract.metagrating_challenge import metagrating
from invrs_gym.challenges.diffract.splitter_challenge import diffractive_splitter
from invrs_gym.challenges.extractor.challenge import photon_extractor
from invrs_gym.challenges.library.challenge import meta_atom_library
from invrs_gym.challenges.metalens.challenge import metalens

BY_NAME = {
    "bayer_sorter": bayer_sorter,
    "ceviche_beam_splitter": ceviche_beam_splitter,
    "ceviche_mode_converter": ceviche_mode_converter,
    "ceviche_power_splitter": ceviche_power_splitter,
    "ceviche_waveguide_bend": ceviche_waveguide_bend,
    "ceviche_wdm": ceviche_wdm,
    "ceviche_lightweight_beam_splitter": ceviche_lightweight_beam_splitter,
    "ceviche_lightweight_mode_converter": ceviche_lightweight_mode_converter,
    "ceviche_lightweight_power_splitter": ceviche_lightweight_power_splitter,
    "ceviche_lightweight_waveguide_bend": ceviche_lightweight_waveguide_bend,
    "ceviche_lightweight_wdm": ceviche_lightweight_wdm,
    "metagrating": metagrating,
    "metalens": metalens,
    "meta_atom_library": meta_atom_library,
    "diffractive_splitter": diffractive_splitter,
    "photon_extractor": photon_extractor,
}



"""Defines metrics used across challenges.

Copyright (c) 2023 The INVRS-IO authors.
"""

from typing import Any, Optional

import jax.numpy as jnp
from jax import lax
from jax import tree_util
from totypes import types

PyTree = Any


def binarization_degree(params: PyTree) -> Optional[jnp.ndarray]:
    """Compute binarization degree, the most sigificant deviation from binary.

    The binarization degree for each density in `params` is computed by doubling
    the most significant deviation from binary of any pixel, and then subtracting
    form one. Thus, for arrays having values within the declared density bounds,
    the minimum possible binarization degree is 0.0, and the maximum is 1.0
    (corresponding to a fully binary design).

    Note that intermediate-valued pixels at the interface between solid and void
    regions are ignored.

    Args:
        params: The params for which the binarization degree is sought.

    Returns:
        The binarization degree, or `None` if `params` contains no densities.
    """
    leaves = tree_util.tree_leaves(
        params, is_leaf=lambda x: isinstance(x, types.Density2DArray)
    )
    densities = [leaf for leaf in leaves if isinstance(leaf, types.Density2DArray)]

    if not densities:
        return None

    def _compute_degree(d: types.Density2DArray) -> jnp.ndarray:
        array = (d.array - d.lower_bound) / (d.upper_bound - d.lower_bound)
        return _array_binarization_degree(jnp.asarray(array))

    degrees = [_compute_degree(d) for d in densities]
    return jnp.amax(jnp.asarray(degrees))


def _array_binarization_degree(x: jnp.ndarray) -> jnp.ndarray:
    """Compute the binarization degree for an array.

    Args:
        x: The array for which the binarization degree is sought.

    Returns:
        The scalar binarization degree.
    """
    x = x.reshape((-1,) + x.shape[-2:])

    violation_lo = jnp.abs(x)
    violation_hi = jnp.abs(1 - x)
    violation = jnp.minimum(violation_lo, violation_hi)

    # Interface pixels must be surrounded by at least four solid or void pixels.
    solid_or_void_pixels = (violation == 0).astype(float)
    solid_or_void_pixels = jnp.pad(
        solid_or_void_pixels,
        ((0, 0), (1, 1), (1, 1)),
        constant_values=1,
    )
    neighbors = lax.conv_general_dilated(
        lhs=solid_or_void_pixels[:, jnp.newaxis, :, :],
        rhs=jnp.ones((1, 1, 3, 3)),
        padding="VALID",
        dimension_numbers=("NCHW", "OIHW", "NCHW"),
        window_strides=(1, 1),
    )
    neighbors = jnp.squeeze(neighbors, axis=1)
    interface_pixels = neighbors >= 4

    violation = jnp.where(interface_pixels, 0.0, violation)
    degree = 1 - 2 * jnp.amax(violation)
    return degree



"""Defines the bayer color sorter challenge.

Copyright (c) 2023 The INVRS-IO authors.
"""

import dataclasses
import functools
from typing import Sequence, Tuple

from fmmax import basis, fmm  # type: ignore[import-untyped]
from invrs_gym import utils
from invrs_gym.challenges import base
from jax import nn
from jax import numpy as jnp
from totypes import symmetry, types

from invrs_gym.challenges.bayer import component as bayer_component
from invrs_gym.utils import materials

EFFICIENCY_BLUE_MEAN = "efficiency_blue_mean"
EFFICIENCY_GREEN_MEAN = "efficiency_green_mean"
EFFICIENCY_RED_MEAN = "efficiency_red_mean"
CROSSTALK_BLUE_MEAN = "crosstalk_blue_mean"
CROSSTALK_GREEN_MEAN = "crosstalk_green_mean"
CROSSTALK_RED_MEAN = "crosstalk_red_mean"

# Cutoff between UV/blue wavelengths, blue/green wavelengths, etc. in microns.
UV_BLUE_CUTOFF = 0.4
BLUE_GREEN_CUTOFF = 0.5
GREEN_RED_CUTOFF = 0.6
RED_IR_CUTOFF = 0.7


density_initializer = functools.partial(
    utils.initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class BayerChallenge(base.Challenge):
    """Defines the bayer color sorter challenge."""

    component: bayer_component.BayerComponent

    def loss(
        self,
        response: bayer_component.BayerResponse,
        efficiency_target: float = 0.55,
    ) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`.

        Args:
            response: The response obtained from simulating the component.
            efficiency_target: Value used in a nonlinearity within the loss function.
                When the efficiency value is far from the target, the dependence of
                loss on efficiency is larger.

        Returns:
            The scalar loss value.
        """
        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)

        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)

        denom = 1 - efficiency_target
        loss_blue = nn.softplus((efficiency_target - efficiency_blue) / denom)
        loss_green = nn.softplus((efficiency_target - efficiency_green) / denom)
        loss_red = nn.softplus((efficiency_target - efficiency_red) / denom)
        return loss_blue + loss_green + loss_red

    def eval_metric(self, response: bayer_component.BayerResponse) -> jnp.ndarray:
        """Compute eval metric from the component `response`.

        The eval metric for the bayer sorter is the minimum of the red, green, and
        blue efficiencies, where the efficiency fo a given color is the average
        efficiency with which wavelengths associated with the color are coupled into
        the appropriate subpixel(s).

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)
        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)
        return jnp.amin(
            jnp.asarray([efficiency_blue, efficiency_green, efficiency_red])
        )

    def metrics(
        self,
        response: bayer_component.BayerResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        metrics = super().metrics(response, params, aux)

        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)
        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)

        # Crosstalk for a given color is total unwanted power in that color's pixel.
        crosstalk_blue = jnp.sum(t_blue * ~is_blue) / jnp.sum(~is_blue)
        crosstalk_green = jnp.sum(t_green * ~is_green) / jnp.sum(~is_green)
        crosstalk_red = jnp.sum(t_red * ~is_red) / jnp.sum(~is_red)

        transmitted_power = aux[bayer_component.TRANSMITTED_POWER]
        assert response.reflection.shape == transmitted_power.shape

        transmitted_power = jnp.sum(response.transmission, axis=-1)
        power = transmitted_power + response.reflection
        assert power.shape == response.wavelength.shape + (2,)

        metrics.update(
            {
                EFFICIENCY_BLUE_MEAN: efficiency_blue,
                EFFICIENCY_GREEN_MEAN: efficiency_green,
                EFFICIENCY_RED_MEAN: efficiency_red,
                CROSSTALK_BLUE_MEAN: crosstalk_blue,
                CROSSTALK_GREEN_MEAN: crosstalk_green,
                CROSSTALK_RED_MEAN: crosstalk_red,
            }
        )
        return metrics


def _detect_color(
    wavelength: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Detects color (blue, green, red) from `wavelength`."""
    is_blue = (wavelength >= UV_BLUE_CUTOFF) & (wavelength < BLUE_GREEN_CUTOFF)
    is_green = (wavelength >= BLUE_GREEN_CUTOFF) & (wavelength < GREEN_RED_CUTOFF)
    is_red = (wavelength >= GREEN_RED_CUTOFF) & (wavelength < RED_IR_CUTOFF)
    return is_blue, is_green, is_red


def _transmission_per_pixel(
    transmission: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes transmission into the blue, green, and red pixels."""
    assert transmission.shape[-1] == 4
    blue_transmission = transmission[..., 0]
    green_transmission = transmission[..., 1] + transmission[..., 2]
    red_transmission = transmission[..., 3]
    return blue_transmission, green_transmission, red_transmission


SYMMETRIES = (symmetry.REFLECTION_NW_SE,)

# Minimum width and spacing are 80 nm for the default dimensions.
MINIMUM_WIDTH = 8
MINIMUM_SPACING = 8


BAYER_SPEC = bayer_component.BayerSpec(
    material_ambient=materials.VACUUM,
    material_metasurface_solid=materials.SI3N4,
    material_metasurface_void=materials.VACUUM,
    material_substrate=materials.SIO2,
    thickness_ambient=1.0,
    thickness_metasurface=types.BoundedArray(0.6, lower_bound=0.4, upper_bound=0.8),
    thickness_substrate=3.0,
    pixel_size=1.0,
    grid_spacing=0.01,
    offset_monitor_substrate=types.BoundedArray(2.4, lower_bound=2.0, upper_bound=2.5),
)

BAYER_SIM_PARAMS = bayer_component.BayerSimParams(
    wavelength=jnp.array([0.45, 0.55, 0.65]),
    polar_angle=0.0,
    azimuthal_angle=0.0,
    formulation=fmm.Formulation.JONES_DIRECT_FOURIER,
    approximate_num_terms=600,
    truncation=basis.Truncation.CIRCULAR,
)


def bayer_sorter(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    thickness_initializer: bayer_component.ThicknessInitializer = (
        utils.initializers.identity_initializer
    ),
    density_initializer: base.DensityInitializer = density_initializer,
    spec: bayer_component.BayerSpec = BAYER_SPEC,
    sim_params: bayer_component.BayerSimParams = BAYER_SIM_PARAMS,
    symmetries: Sequence[str] = SYMMETRIES,
) -> BayerChallenge:
    """The bayer color sorter challenge.

    The bayer sorter challenge entails the design of a color-sorting metasurface that
    replaces the color filter array in a conventional image sensor. The thickness of
    the metasurface, and the distance between metasurface and focal plane are further
    degrees of freedom. The bayer challenge is based on "Pixel-level Bayer-type colour
    router based on metasurfaces" by Zou et al.

    https://www.nature.com/articles/s41467-022-31019-7

    Args:
        minimum_width: The minimum width target for the challenge, in pixels.  The
            default value of 8 corresponds to a physical size of approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        thickness_initializer: Callable which returns the initial thickness, given a
            key and seed thickness.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        spec: Defines the physical specification of the bayer sorter.
        sim_params: Defines the simulation settings of the bayer sorter.
        symmetries: Defines the symmetries of the bayer sorter.

    Returns:
        The `BayerChallenge`.
    """
    return BayerChallenge(
        component=bayer_component.BayerComponent(
            spec=spec,
            sim_params=sim_params,
            thickness_initializer=thickness_initializer,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
    )
